from airflow.models.baseoperator import BaseOperator
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
# from airflow.utils.decorators import apply_defaults
from airflow.exceptions import AirflowException

from gcp_airflow_foundations.operators.api.hooks.mailgun_hook_test import MailgunHook

import logging
import os
import requests
import zipfile


def get_bigquery_hook(bq_conn: str):
    try:
        hook = BigQueryHook(bq_conn, use_legacy_sql=False)
        logging.info("BigQuery connection established")
        return hook
    except Exception as e:
        raise AirflowException(f"Connection not established: {str(e)}")


class BigQueryToCsv(BaseOperator):
    # @apply_defaults
    def __init__(self, *, bq_conn: str, query: str, file: str, **kwargs):
        super().__init__(**kwargs)

        self.bq_hook = get_bigquery_hook(bq_conn)
        self.query = query
        self.file = file

    def execute(self, context):
        # Grab phone_number column from BigQuery source table as a pandas dataframe
        df = self.bq_hook.get_pandas_df(self.query)

        # Check for directory in local FS. If it doesn't exist, create one.
        outdir = "/tmp/mailgun"
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        # save dataframe to local system directory
        df.to_csv(os.path.join(outdir, self.file), index=False, header=True)


class CsvToMailgunToCsv(BaseOperator):
    # @apply_defaults
    def __init__(self, *, mailgun_conn: str, source_file: str, job_id: str, **kwargs):
        super().__init__(**kwargs)
        self.hook = MailgunHook(mailgun_conn_id=mailgun_conn)
        self.source_file = source_file
        self.job_id = job_id

    def execute(self, context):
        res = self.hook.validate_email_bulk(self.source_file, self.job_id)
        logging.info(f"Mailgun job sent. Response status: {res}")


class DownloadCSV(BaseOperator):
    # @apply_defaults
    def __init__(self, *, mailgun_conn: str, job_id: str, target_file: str, **kwargs):
        super().__init__(**kwargs)
        self.hook = MailgunHook(mailgun_conn_id=mailgun_conn)
        self.job_id = job_id
        self.target_file = target_file

    def execute(self, context):
        # get link for download
        link = self.hook.download_csv(self.job_id)
        print(link)

        # download file
        outdir = "/tmp/mailgun"
        if not os.path.exists(outdir):
            os.mkdir(outdir)

        csv_zip = requests.get(link)
        with open(os.path.join(outdir, self.target_file) + ".zip", "wb") as f:
            f.write(csv_zip.content)

        # unzip file to csv
        with zipfile.ZipFile(
            os.path.join(outdir, self.target_file) + ".zip", "r"
        ) as zip_ref:
            zip_ref.extractall(os.path.join(outdir, "validated"))



from airflow import DAG

from airflow.utils.task_group import TaskGroup
from airflow.operators.python import PythonOperator

from typing import List, Optional

from gcp_airflow_foundations.operators.gcp.ods.ods_merge_table_operator import (
    MergeBigQueryODS
)
from gcp_airflow_foundations.operators.gcp.schema_migration.schema_migration_operator import (
    MigrateSchema,
)

from gcp_airflow_foundations.operators.gcp.create_table import (
    CustomBigQueryCreateEmptyTableOperator
)
from gcp_airflow_foundations.enums.ingestion_type import IngestionType

from gcp_airflow_foundations.base_class.ods_table_config import OdsTableConfig

from gcp_airflow_foundations.operators.gcp.create_dataset import CustomBigQueryCreateEmptyDatasetOperator


def ods_builder(
    project_id: str,
    table_id: str,
    dag_table_id: str,
    dataset_id: str,
    landing_zone_dataset: str,
    landing_zone_table_name_override: Optional[str],
    column_mapping: Optional[dict],
    column_casting: Optional[dict],
    new_column_udfs: Optional[dict],
    surrogate_keys: List[str],
    ingestion_type: IngestionType,
    ods_table_config: Optional[OdsTableConfig],
    partition_expiration: Optional[int],
    location: str,
    dag: DAG,
    cluster_fields=None,
    labels=None,
    encryption_configuration=None,
) -> TaskGroup:

    """
    Method for returning a Task Group for 1) creating an empty target ODS table (if it doesn't already exist) and for 2) for merging the staging table data into the target ODS table
    """
    taskgroup = TaskGroup(group_id="create_ods_merge_taskgroup")

    if ods_table_config.ods_table_time_partitioning is not None:
        field = (
            column_mapping.get(
                ods_table_config.partition_column_name,
                ods_table_config.partition_column_name,
            )
            if column_mapping
            else ods_table_config.partition_column_name
        )

        time_partitioning = {
            "type": ods_table_config.ods_table_time_partitioning.value,
            "field": field,
            "expirationMs": partition_expiration,
        }

    else:
        time_partitioning = None

    ods_task = PythonOperator(
        task_id="run_ods_tasks",
        op_kwargs={
            "project_id": project_id,
            "dataset_id": dataset_id,
            "location": location,
            "taskgroup": taskgroup,
            "table_id": table_id,
            "dag_table_id": dag_table_id,
            "cluster_fields": cluster_fields,
            "time_partitioning": time_partitioning,
            "dag": dag,
            "landing_zone_dataset": landing_zone_dataset,
            "landing_zone_table_name_override": landing_zone_table_name_override,
            "surrogate_keys": surrogate_keys,
            "column_mapping": column_mapping,
            "column_casting": column_casting,
            "new_column_udfs": new_column_udfs,
            "ingestion_type": ingestion_type,
            "ods_table_config": ods_table_config
        },
        python_callable=ods_task_operator,
        task_group=taskgroup,
        dag=dag
    ) 

    ods_task

    return taskgroup

def ods_task_operator(
    project_id,
    dataset_id,
    location,
    taskgroup,
    table_id,
    dag_table_id,
    cluster_fields,
    time_partitioning,
    landing_zone_dataset,
    landing_zone_table_name_override,
    surrogate_keys,
    column_mapping,
    column_casting,
    new_column_udfs,
    ingestion_type,
    ods_table_config,
    **kwargs
):

    create_dataset = CustomBigQueryCreateEmptyDatasetOperator(
        task_id="create_ods_dataset",
        project_id=project_id,
        dataset_id=dataset_id,
        location=location,
        exists_ok=True,
        task_group=taskgroup,
    )
    create_dataset.execute(context=kwargs)

    # 1 Check if ODS table exists and if not create an empty table
    create_table = CustomBigQueryCreateEmptyTableOperator(
        task_id="create_ods_table",
        project_id=project_id,
        dataset_id=dataset_id,
        table_id=table_id,
        dag_table_id=dag_table_id,
        cluster_fields=cluster_fields,
        time_partitioning=time_partitioning,
        task_group=taskgroup,
    )
    create_table.execute(context=kwargs)

    # 2 Migrate schema
    migrate_schema = MigrateSchema(
        task_id="schema_migration",
        project_id=project_id,
        table_id=table_id,
        dag_table_id=dag_table_id,
        dataset_id=dataset_id,
        task_group=taskgroup,
    )
    migrate_schema.execute(context=kwargs)

    # 3 Merge or truncate tables based on the ingestion type defined in the config file and insert metadata columns
    insert = MergeBigQueryODS(
        task_id=f"upsert_{table_id}",
        project_id=project_id,
        stg_dataset_name=landing_zone_dataset,
        data_dataset_name=dataset_id,
        stg_table_name=landing_zone_table_name_override,
        data_table_name=table_id,
        dag_table_id=dag_table_id,
        surrogate_keys=surrogate_keys,
        column_mapping=column_mapping,
        column_casting=column_casting,
        new_column_udfs=new_column_udfs,
        ingestion_type=ingestion_type,
        ods_table_config=ods_table_config,
        location=location,
        task_group=taskgroup,
    )
    insert.execute(context=kwargs)


import logging
from airflow.models import BaseOperator
from airflow.operators.python import PythonOperator
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryDeleteTableOperator,
)
from airflow.providers.google.cloud.operators.dlp import CloudDLPCreateDLPJobOperator
from airflow.utils.task_group import TaskGroup
from google.cloud.bigquery.dataset import DatasetReference
from google.cloud.bigquery.table import TableReference

from gcp_airflow_foundations.base_class.dlp_table_config import DlpTableConfig
from gcp_airflow_foundations.operators.branch.branch_on_cron_operator import (
    BranchOnCronOperator,
)
from gcp_airflow_foundations.operators.gcp.dlp.dlp_helpers import (
    results_to_bq_policy_tags,
)
from gcp_airflow_foundations.operators.gcp.dlp.dlp_job_helpers import (
    build_inspect_job_config,
)
from gcp_airflow_foundations.operators.gcp.dlp.get_dlp_bq_inspection_results_operator import (
    DlpBQInspectionResultsOperator,
)


def dlp_policy_tag_taskgroup_name():
    return "dlp_policy_tags"


def schedule_dlp_to_datacatalog_taskgroup(
    datastore: str,
    project_id: str,
    table_id: str,
    dataset_id: str,
    table_dlp_config: DlpTableConfig,
    next_task: BaseOperator,
    dag,
):
    """
    We want to schedule DLP to run
    1) First time the table is ingested
    2) On schedule there after
    """

    taskgroup = TaskGroup(dlp_policy_tag_taskgroup_name(), dag=dag)

    dlp_task_group = dlp_to_datacatalog_builder(
        taskgroup=taskgroup,
        datastore=datastore,
        project_id=project_id,
        table_id=table_id,
        dataset_id=dataset_id,
        table_dlp_config=table_dlp_config,
        next_task=next_task,
        dag=dag,
    )

    decide = BranchOnCronOperator(
        task_id="check_if_should_run_dlp",
        follow_task_ids_if_true=dlp_task_group,
        follow_task_ids_if_false=next_task.task_id,
        cron_expression=table_dlp_config.source_config.recurrence_schedule,
        use_task_execution_day=True,
        run_on_first_execution=table_dlp_config.source_config.run_on_first_execution,
        dag=dag,
    )

    decide >> dlp_task_group
    return decide


def schedule_dlp_to_datacatalog_taskgroup_multiple_tables(
    table_configs: list, table_dlp_config: DlpTableConfig, next_task: BaseOperator, dag
):
    """
    Check if DLP should run, and run it on multiple tables
    We want to schedule DLP to run
    1) First time the table is ingested
    2) On schedule there after
    """
    logging.info("schedule_dlp_to_datacatalog_taskgroup_multiple_tables")

    taskgroup = TaskGroup(dlp_policy_tag_taskgroup_name(), dag=dag)

    dlp_tasks = []
    for table_config in table_configs:
        dlp_task = dlp_to_datacatalog_builder(
            taskgroup=taskgroup,
            datastore=table_config["datastore"],
            project_id=table_config["project_id"],
            table_id=table_config["table_id"],
            dataset_id=table_config["dataset_id"],
            table_dlp_config=table_dlp_config,
            next_task=next_task,
            dag=dag,
        )
        dlp_tasks.append(dlp_task)

    dlp_task_ids = list(map(lambda t: t.task_id, dlp_tasks))

    logging.info(
        f"Run BranchOnCronOperator for  {dlp_task_ids} next_task {next_task.task_id}"
    )

    decide = BranchOnCronOperator(
        task_id="check_if_should_run_dlp",
        follow_task_ids_if_true=dlp_task_ids,
        follow_task_ids_if_false=next_task.task_id,
        cron_expression=table_dlp_config.source_config.recurrence_schedule,
        use_task_execution_day=True,
        run_on_first_execution=table_dlp_config.source_config.run_on_first_execution,
        task_group=taskgroup,
        dag=dag,
    )

    decide >> dlp_tasks
    return decide


def dlp_to_datacatalog_builder(
    taskgroup: TaskGroup,
    datastore: str,
    project_id: str,
    table_id: str,
    dataset_id: str,
    table_dlp_config: DlpTableConfig,
    next_task: BaseOperator,
    dag,
) -> TaskGroup:
    """
  Method for returning a Task Group for scannign a table with DLP, and creating BigQuery policy tags based on the results
   1) Scan table with DLP and write results to BigQuery
   2) Schedule future DLP
   3) Read results of DLP scan from BigQuery
   4) Update Policy Tags in BQ
   Returns the first task
  """

    assert table_dlp_config.source_config is not None

    # setup tables vars
    dlp_results_dataset_id = table_dlp_config.source_config.results_dataset_id

    table_ref = TableReference(DatasetReference(project_id, dataset_id), table_id)
    dlp_results_table_ref = TableReference(
        DatasetReference(project_id, dlp_results_dataset_id), f"{table_id}_dlp_results"
    )
    dlp_results_table = f"{dlp_results_table_ref.project}.{dlp_results_table_ref.dataset_id}.{dlp_results_table_ref.table_id}"

    # setup DLP scan vars
    dlp_template_name = table_dlp_config.get_template_name()
    rows_limit_percent = table_dlp_config.get_rows_limit_percent()

    inspect_job = build_inspect_job_config(
        dlp_template_name, table_ref, rows_limit_percent, dlp_results_table_ref
    )

    # 1 First delete the results table
    delete_dlp_results = BigQueryDeleteTableOperator(
        task_id=f"delete_old_dlp_results_{datastore}",
        deletion_dataset_table=dlp_results_table,
        ignore_if_missing=True,
        task_group=taskgroup,
        dag=dag,
    )

    # 2 Scan table
    scan_task = CloudDLPCreateDLPJobOperator(
        task_id=f"scan_table_{datastore}",
        project_id=project_id,
        inspect_job=inspect_job,
        wait_until_finished=True,
        task_group=taskgroup,
        dag=dag,
    )
    # 4. Read results
    read_results_task = DlpBQInspectionResultsOperator(
        task_id=f"read_dlp_results_{datastore}",
        project_id=dlp_results_table_ref.project,
        dataset_id=dlp_results_table_ref.dataset_id,
        table_id=dlp_results_table_ref.table_id,
        do_xcom_push=True,
        min_match_count=table_dlp_config.get_min_match_count(),
        task_group=taskgroup,
        dag=dag,
    )

    # 5. Update policy tags
    update_tags_task = PythonOperator(
        task_id=f"update_bq_policy_tags_{datastore}",
        python_callable=update_bq_policy_tags,  # <--- PYTHON LIBRARY THAT COPIES FILES FROM SRC TO DEST
        task_group=taskgroup,
        dag=dag,
        templates_dict={
            "dlp_results": f"{{{{ti.xcom_pull(task_ids='{read_results_task.task_id}')}}}}",
            # "dlp_results": "{{ti.xcom_pull(task_ids='dlp_policy_tags.read_dlp_results_test')}}",
        },
        op_kwargs={
            "project_id": project_id,
            "dataset_id": table_ref.dataset_id,
            "table_id": table_ref.table_id,
            "policy_tag_config": table_dlp_config.source_config.policy_tag_config,
            "task_ids": read_results_task.task_id,
        },
        provide_context=True,
    )

    delete_dlp_results >> scan_task >> read_results_task >> update_tags_task >> next_task

    return delete_dlp_results


def update_bq_policy_tags(
    project_id,
    dataset_id,
    table_id,
    policy_tag_config,
    gcp_conn_id="google_cloud_default",
    delegate_to=None,
    **context,
):
    dlp_results = context["templates_dict"]["dlp_results"]

    hook = BigQueryHook(gcp_conn_id=gcp_conn_id, delegate_to=delegate_to)

    schema_fields_updates = results_to_bq_policy_tags(
        project_id, dlp_results, policy_tag_config
    )

    logging.info(
        f"updating policy tags for {project_id}.{dataset_id}.{table_id} to {schema_fields_updates}"
    )
    hook.update_table_schema(
        dataset_id=dataset_id,
        table_id=table_id,
        schema_fields_updates=schema_fields_updates,
        include_policy_tags=False  # Seem to work fine with False.
        # When set to True it tries to add a policy tags instead of replacing
    )



from enum import Enum, unique


@unique
class Level(Enum):
    AD = "ad"
    ADSET = "adset"
    CAMPAIGN = "campaign"
    ACCOUNT = "account"


@unique
class DatePreset(Enum):
    TODAY = "today"
    YESTERDAY = "yesterday"
    THIS_MONTH = "this_month"
    LAST_MONTH = "last_month"
    MAXIMUM = "maximum"


@unique
class AccountLookupScope(Enum):
    ALL = "all"
    ACTIVE = "active"


@unique
class ApiObject(Enum):
    INSIGHTS = "INSIGHTS"
    CAMPAIGNS = "CAMPAIGNS"
    ADSETS = "ADSETS"



import os
import tempfile
import logging
from typing import Dict, Optional

from airflow.models import BaseOperator
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.salesforce.hooks.salesforce import SalesforceHook
from airflow.providers.google.cloud.transfers.salesforce_to_gcs import (
    SalesforceToGcsOperator,
)

from airflow import AirflowException


class SalesforceToGcsQueryOperator(SalesforceToGcsOperator):
    """
    Generates a Salesforce query, submits it, and uploads results to Google Cloud Storage
    .. seealso::
        Extends the SalesforceToGcsOperator, see the link
        :ref:`howto/operator:SalesforceToGcsOperator`
    :param salesforce_object: The object to query in Salesforce
    :type salesforce_object: str
    :param ingest_all_fields: Whether or not to Select * the Object in Salesforce (if True,
        then parameters fields_to_omit and fields_to_ingest are ignored)
    :type ingest_all_fields:  bool
    :param fields_to_omit: A list of fields to omit from the query (if included,
        then parameters ingest_all_fields and fields_to_omit are ignored)
    :type fields_to_omit: List[str]
    :param fields_to_ingest: A list of fields to include in the query (if included,
        then parameters ingest_all_fields and fields_to_omit are ignored)
    :type fields_to_ingest: List[str]
    :param bucket_name: The bucket to upload to.
    :type bucket_name: str
    :param object_name: The object name to set when uploading the file.
    :type object_name: str
    :param salesforce_conn_id: the name of the connection that has the parameters
        we need to connect to Salesforce.
    :type salesforce_conn_id: str
    :param include_deleted: True if the query should include deleted records.
    :type include_deleted: bool
    :param query_params: Additional optional arguments
    :type query_params: dict
    :param export_format: Desired format of files to be exported.
    :type export_format: str
    :param coerce_to_timestamp: True if you want all datetime fields to be converted into Unix timestamps.
        False if you want them to be left in the same format as they were in Salesforce.
        Leaving the value as False will result in datetimes being strings. Default: False
    :type coerce_to_timestamp: bool
    :param record_time_added: True if you want to add a Unix timestamp field
        to the resulting data that marks when the data was fetched from Salesforce. Default: False
    :type record_time_added: bool
    :param gzip: Option to compress local file or file data for upload
    :type gzip: bool
    :param gcp_conn_id: the name of the connection that has the parameters we need to connect to GCS.
    :type gcp_conn_id: str
    """

    template_fields = ("ingest_all_fields", "fields_to_omit", "fields_to_include")
    template_ext = (".sql",)

    def __init__(
        self,
        *,
        salesforce_object: str,
        ingest_all_fields: bool,
        fields_to_omit,
        fields_to_include,
        **kwargs,
    ):
        super().__init__(query="", **kwargs)
        self.salesforce_object = salesforce_object
        self.ingest_all_fields = ingest_all_fields
        self.fields_to_omit = fields_to_omit
        self.fields_to_include = fields_to_include

    def execute(self, context: Dict):
        fields_to_query = []

        sf_hook = SalesforceHook(salesforce_conn_id=self.salesforce_conn_id)
        fields = sf_hook.get_available_fields(obj=self.salesforce_object)
        logging.info(fields)

        if self.ingest_all_fields:
            # Query all fields
            fields_to_query = fields
        elif self.fields_to_omit:
            # Query all fields except [x_1,....x_n]
            fields_to_query = list(set(fields) - set(self.fields_to_omit))
        elif self.fields_to_include:
            # Query specified fields
            fields_to_query = self.fields_to_include
            if not set(fields_to_query).issubset(set(fields)):
                extra_fields = list(set(fields_to_query) - set(fields))
                raise AirflowException(
                    f"Queryable content for Salesforce object {self.salesforce_object} included fields not present in the object's schema: {extra_fields}"
                )
        else:
            # ingest_all_fields is false, and no other information was provided - throw exception
            raise AirflowException(
                f"Queryable content for Salesforce object {self.salesforce_object} was malformed"
            )
        query = f"""SELECT {",".join(fields_to_query)} FROM {self.salesforce_object}""".strip()
        logging.info(query)
        self.query = query

        super().execute(context)



import logging


def get_query_for_oracle_load_full(table_name, columns, owner):
    """
    JDBC query for full ingestion of one table
    """

    logging.info(f"BUILDING FULL QUERY for {table_name}")

    select_cols = ",".join(str(x) for x in columns)

    return f"select {select_cols} from {owner}.{table_name}"


def get_query_for_oracle_load_incremental(table_name, columns, date_column, owner, ds, num_backtrack_days=1):
    """
    JDBC query for full ingestion of one table
    """

    logging.info(f"BUILDING INCREMENTAL QUERY for {table_name}")

    select_cols = ",".join(str(x) for x in columns)

    if num_backtrack_days == 1:
        where_clause = f"where {date_column} = to_date('{ds}', 'YYYY-MM-DD')"
    else:
        where_clause = f"where {date_column} = to_date('{ds}', 'YYYY-MM-DD')- {num_backtrack_days - 1}"

    return f'select {select_cols} from {owner}.{table_name} {where_clause}'


def convert_schema_to_json(lists, labels):
    """
    Input:
        lists: non-empty list of n lists each of length x   2
        labels: list of strings of length n
    Output:
        list of x dictionaries with n entries, each row corresponding
        to a labelled row (merged from lists)
    """
    dicts = []
    for i in range(len(lists[0])):
        dict = {}
        for j in range(len(labels)):
            dict[labels[j]] = lists[j][i]
        dicts.append(dict)
    return dicts


def cast_columns(columns, dtypes, casts):
    # perform the Oracle castings needed within the query itself.
    castings = []
    for c, d in zip(columns, dtypes):
        if d in casts:
            cast = casts[d].replace("COLUMN", c)
            castings.append(cast)
        else:
            castings.append(c)
    return castings


def get_schema_query(owner):
    """
    Oracle query for getting schema information for all tables
    """

    cols = ["TABLE_NAME", "COLUMN_NAME", "DATA_TYPE", "COLUMN_ID"]
    cols = ",".join(cols)

    return f"""select {cols} from alL_tab_columns where OWNER = '{owner}'"""


def get_table_schema_query(schema_table, source_table_name):
    """
    Query to get source schema for table from BQ (previously ingested).
    """

    return f" select distinct * from (select COLUMN_NAME, DATA_TYPE from `{schema_table}` \
        where TABLE_NAME = '{source_table_name}' order by COLUMN_ID)"


def oracle_mappings():
    return {
        # STRING
        "VARCHAR2": "STRING",
        "NVARCHAR2": "STRING",
        "CHAR": "STRING",
        "NCHAR": "STRING",
        "CLOB": "STRING",
        "NCLOB": "STRING",
        "INTERVAL YEAR TO MONTH": "STRING",
        "INTERVAL DAY TO SECOND": "STRING",
        "BFILE": "STRING",
        "ROWID": "STRING",
        # INT64
        "INTEGER": "INT64",
        "SHORTINTEGER": "INT64",
        "LONGINTEGER": "INT64",
        "NUMBER(x, -y)": "INT64",
        "NUMBER(x)": "INT64",
        # NUMERIC
        "NUMBER": "NUMERIC",
        "FLOAT": "NUMERIC",
        "BINARY_DOUBLE": "NUMERIC",
        "BINARY_FLOAT": "NUMERIC",
        "NUMBER(*, x)": "NUMERIC",
        # BYTES
        "LONG": "BYTES",
        "BLOB": "BYTES",
        "RAW": "BYTES",
        "LONG RAW": "BYTES",
        # DATE
        "DATE": "DATE",
        # TIMESTAMP
        "TIMESTAMP": "TIMESTAMP",
        "TIMESTAMP()": "TIMESTAMP",
        "TIMESTAMP WITH TIMEZONE": "TIMESTAMP",
        "TIMESTAMP WITH LOCAL TIME ZONE": "TIMESTAMP",
    }


def oracle_to_bq(dtypes: list):
    mappings = oracle_mappings()
    # Dicts are not really intended for use in this way, but for a tiny one
    # this is just easier
    for i in range(len(dtypes)):
        dtypes[i] = mappings[dtypes[i]]
    return dtypes



from xxlimited import Str
from dacite import Config
from pydantic import validator, root_validator
from pydantic.dataclasses import dataclass
from dataclasses import field

from typing import List, Optional
import regex as re

from gcp_airflow_foundations.enums.template_ingestion import TemplateIngestionMode, TemplateDagCreationMode


@dataclass
class TemplateIngestionOptionsConfig:
    """
    Attributes:
        ingest_all_tables: if true, ingest all tables from source
        ingestion_name: name to use for DAG id, should correspond to the relevant partition of the source cut out by the regex expression provided
        dag_creation_mode: if "SOURCE", then one DAG per source is created. if "TABLE", one DAG per table.
        regex_table_matching: regex pattern to match tables to, if ingest_all_tables is false
    """
    ingestion_name: str
    ingest_mode: TemplateIngestionMode = TemplateIngestionMode("INGEST_BY_TABLE_NAMES"),
    dag_creation_mode: TemplateDagCreationMode = TemplateDagCreationMode("TABLE"),
    table_names: list = field(default_factory=list)
    regex_pattern: Optional[str] = None



import logging


class SqlHelperODS:
    """
    SQL helper class used to formulate the SQL queries for the merge operations of ODS tables.

    :param source: Source table name
    :type source: str
    :param target: Target table name
    :type target: str
    :param source_dataset: Source dataset name
    :type source_dataset: str
    :param target_dataset: Target dataset name
    :type target_dataset: str
    :param surrogate_keys: Surrogate keys of the target table
    :type surrogate_keys: list
    :param columns: List of source table columns
    :type columns: list
    :param gcp_conn_id: Airflow GCP connection ID
    :type gcp_conn_id: str
    :param column_mapping: Column mapping
    :type column_mapping: dict
    :param new_column_udfs: New column UDFs
    :type new_column_udfs: dict
    :param time_partitioning: Time partitioning option for BigQuery target table. One of HOUR, DAY, or MONTH
    :type time_partitioning: str
    :param ods_metadata: User-provided options for ODS metadata column naming
    :type ods_metadata: OdsTableMetadataConfig
    """

    def __init__(
        self,
        source_dataset,
        target_dataset,
        source,
        target,
        surrogate_keys,
        column_mapping,
        column_casting,
        new_column_udfs,
        columns,
        ods_metadata,
        partition_column_name=None,
        time_partitioning=None,
        partition_timestamp=None,
        gcp_conn_id="google_cloud_default",
    ):

        self.source_dataset = source_dataset
        self.target_dataset = target_dataset
        self.source = source
        self.target = target
        self.surrogate_keys = surrogate_keys
        self.column_mapping = column_mapping
        self.column_casting = column_casting
        self.new_column_udfs = new_column_udfs
        self.ods_metadata = ods_metadata
        self.gcp_conn_id = gcp_conn_id
        self.columns = columns
        self.hash_column_name = ods_metadata.hash_column_name
        self.primary_key_hash_column_name = ods_metadata.primary_key_hash_column_name
        self.ingestion_time_column_name = ods_metadata.ingestion_time_column_name
        self.update_time_column_name = ods_metadata.update_time_column_name
        self.partition_column_name = partition_column_name
        self.time_partitioning = time_partitioning
        self.partition_timestamp = partition_timestamp

        if column_casting:
            self.columns_str_source: str = ",".join(
                [
                    column_casting[col]["function"].format(column=col)
                    if col in column_casting
                    else "`{}`".format(col)
                    for col in columns
                ]
            )
        else:
            self.columns_str_source: str = ",".join(
                ["`{}`".format(col) for col in columns]
            )

        if self.new_column_udfs:
            keys = list(self.new_column_udfs.keys())
            self.columns_str_source = self.columns_str_source + "," + \
                ",".join(
                    new_column_udfs[col]["function"]
                    for col in keys
                )

        self.columns_str_keys: str = ",".join(surrogate_keys)
        self.columns_str_target: str = ",".join(
            ["`{}`".format(self.column_mapping[i]) for i in columns]
        )

    def create_insert_sql(self):
        rows = f"""{self.columns_str_target}, {self.ingestion_time_column_name}, {self.update_time_column_name}, {self.hash_column_name}, {self.primary_key_hash_column_name}"""
        values = f"""{self.columns_str_source}, CURRENT_TIMESTAMP(), CURRENT_TIMESTAMP(), TO_BASE64(MD5(TO_JSON_STRING(S))), TO_BASE64(MD5(ARRAY_TO_STRING([{",".join(["CAST(S.`{}` AS STRING)".format(surrogate_key) for surrogate_key in self.surrogate_keys])}], "")))"""
        return rows, values

    def create_full_sql(self):
        if self.column_casting:
            COLUMNS = ",".join(
                f"{self.column_casting[col]['function'].format(column=f'`{col}`')} AS `{self.column_mapping[col]}`"
                if col in self.column_casting
                else f"`{col}` AS `{self.column_mapping[col]}`"
                for col in self.columns
            )

        else:
            COLUMNS = ",".join(
                f"{col} AS `{self.column_mapping[col]}`" for col in self.columns
            )

        if self.new_column_udfs:
            keys = list(self.new_column_udfs.keys())
            COLUMNS = COLUMNS + "," + ",".join(
                f"{self.new_column_udfs[col]['function']} AS `{self.column_mapping[col]}`" for col in keys
            )

        return f"""
            SELECT {COLUMNS},
                CURRENT_TIMESTAMP() AS {self.ingestion_time_column_name},
                CURRENT_TIMESTAMP() AS {self.update_time_column_name},
                TO_BASE64(MD5(TO_JSON_STRING(S))) AS {self.hash_column_name},
                TO_BASE64(MD5(ARRAY_TO_STRING([{",".join(["CAST(S.`{}` AS STRING)".format(surrogate_key) for surrogate_key in self.surrogate_keys])}], ""))) AS {self.primary_key_hash_column_name}
            FROM `{self.source_dataset}.{self.source}` S
        """

    def create_upsert_sql_with_hash(self):
        if self.column_casting:
            UPDATE_COLUMNS = ",".join(
                f"`{self.column_mapping[col]}`={self.column_casting[col]['function'].format(column=f'S.{col}')}"
                if col in self.column_casting
                else f"`{self.column_mapping[col]}`=S.`{col}`"
                for col in self.columns
            )
        else:
            UPDATE_COLUMNS = ",".join(
                f"`{self.column_mapping[col]}`=S.`{col}`" for col in self.columns
            )

        if self.new_column_udfs:
            # keys = list(self.new_column_udfs.keys())
            # COLUMNS = COLUMNS + "," + ",".join(f"{col} AS `{self.column_mapping[col]}`" for col in keys)
            pass

        comma = ","

        rows, values = self.create_insert_sql()

        if self.time_partitioning:
            partition_filter = f"""
                AND
                    T.`{self.column_mapping[self.partition_column_name]}` = TIMESTAMP_TRUNC('{self.partition_timestamp}', {self.time_partitioning})
                AND
                    S.`{self.partition_column_name}` = TIMESTAMP_TRUNC('{self.partition_timestamp}', {self.time_partitioning})
            """
        else:
            partition_filter = ""

        return f"""
            MERGE `{self.target_dataset}.{self.target}` T
            USING `{self.source_dataset}.{self.source}` S
            ON {' AND '.join([f'T.{self.column_mapping[surrogate_key]}=S.{surrogate_key}' for surrogate_key in self.surrogate_keys])} {partition_filter}
            WHEN MATCHED THEN UPDATE
                SET {UPDATE_COLUMNS + f'{comma}' + f'{self.update_time_column_name}=CURRENT_TIMESTAMP()' + f'{comma}' +  f'{self.hash_column_name}=TO_BASE64(MD5(TO_JSON_STRING(S)))' }
            WHEN NOT MATCHED THEN
                INSERT ({rows})
                VALUES ({values})"""



from dataclasses import fields
from os import X_OK
from urllib.parse import urlparse
import logging
from dacite import from_dict
from dataclasses import dataclass
from datetime import datetime
import shutil
import os
import glob

from airflow.operators.python import PythonOperator
from airflow.operators.dummy import DummyOperator
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.sensors.python import PythonSensor
from airflow.models import Variable

from airflow.providers.sftp.hooks.sftp import SFTPHook
from gcp_airflow_foundations.operators.api.sensors.sftp_sensor import SFTPFilesExistenceSensor
from gcp_airflow_foundations.operators.api.sensors.sftp_prefix_sensor import SFTPFilesExistencePrefixSensor
from gcp_airflow_foundations.source_class.generic_file_source import GenericFileIngestionDagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder
from gcp_airflow_foundations.base_class.sftp_table_config import SFTPTableConfig
from gcp_airflow_foundations.common.sftp.helpers import save_id_to_file


class SFTPFileIngestionDagBuilder(GenericFileIngestionDagBuilder):
    """
    Builds DAGs to load files from an SFTP server to a BigQuery Table. Authenticates with the Airflow SFTP provider connection.

    For specifying files to be uploaded, there are 3 options:
        1) A metadata file is provided with templated daily file names (templating supports {{ ds }} and {{ TABLE_NAME }})
        2) The field templated_file_name is provided in the table config (the same templating applies - but only 1 file can be provided in this manner)
        3) A directory can be specified for .PARQUET files only in the table config

    If a metadata file is provided, it can be a fixed file, or can be a new file supplied daily.
    Airflow context variables are supported for the file naming, e.g.
        TABLE_METADATA_FILE_{{ ds }}.csv
    for a metadata file supplied daily.

    The format of the metadata file should be a csv with one column as follows:
        FILE_NAME_1
        ...
        FILE_NAME_N
    with all files to ingest

    If a directory for .PARQUET ingestion is provided, the field directory_prefix should point to the directory where
    the outward-most external partition sits, and the field flag_file_path should point to the path of the _SUCCESS file on the sftp server.
    A mod_time check occurs on the flag file to screen for daily ingestion on the sensor.
    """
    source_type = "SFTP"

    def flag_file_sensor(self, table_config, taskgroup):
        if "flag_file_path" in table_config.extra_options.get("file_table_config"):
            dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]
            flag_file_path = table_config.extra_options.get("file_table_config")["flag_file_path"]
            return self.get_sftp_sensor(table_config, taskgroup, dir_prefix, flag_file_path)
        else:
            return None

    def file_ingestion_task(self, table_config, taskgroup):
        """
        Ingest from SFTP
        """
        dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]
        if "flag_file_path" in table_config.extra_options.get("file_table_config"):
            flag_file_path = table_config.extra_options.get("file_table_config")["flag_file_path"]
        else:
            flag_file_path = ""
        gcs_bucket_prefix = self.config.source.extra_options["file_source_config"]["gcs_bucket_prefix"]
        bucket = self.config.source.extra_options["gcs_bucket"]
        return self.get_sftp_ingestion_operator(table_config, taskgroup, dir_prefix, gcs_bucket_prefix, flag_file_path, bucket)

    def get_file_list_task(self, table_config, taskgroup):
        return PythonOperator(
            task_id="get_file_list",
            op_kwargs={"table_config": table_config},
            python_callable=self.get_list_of_files,
            task_group=taskgroup)

    def get_list_of_files(self, table_config, **kwargs):
        # gcs_hook = GCSHook()
        airflow_date_template = self.config.source.extra_options["file_source_config"]["airflow_date_template"]
        if airflow_date_template == "ds":
            ds = kwargs["ds"]
        else:
            ds = kwargs["prev_ds"]
        ds = datetime.strptime(ds, "%Y-%m-%d").strftime(self.config.source.extra_options["file_source_config"]["date_format"])
        # XCom push the list of files
        # overwrite if in table_config
        dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]
        dir_prefix = dir_prefix.replace("{{ ds }}", ds)

        # gcs_bucket_prefix = self.config.source.extra_options["file_source_config"]["gcs_bucket_prefix"]

        if self.config.source.extra_options["file_source_config"]["source_format"] == "PARQUET":
            file_list = [dir_prefix]
            kwargs['ti'].xcom_push(key='file_list', value=file_list)
            return
        else:
            # bucket = self.config.source.extra_options["gcs_bucket"]
            if "metadata_file" in table_config.extra_options.get("file_table_config"):
                # metadata_file_name = table_config.extra_options.get("file_table_config")["metadata_file"]
                # metadata_file = gcs_hook.download(bucket_name=bucket, object_name=metadata_file_name, filename="metadata.csv")
                file_list = []
                with open('metadata.csv', newline='') as f:
                    for line in f:
                        file_list.append(line.strip())
            else:
                templated_file_name = self.config.source.extra_options["file_source_config"]["file_name_template"]
                templated_file_name = templated_file_name.replace("{{ TABLE_NAME }}", table_config.table_name)
                file_list = [templated_file_name]

            # support replacing files with current dates
            file_list[:] = [file.replace("{{ ds }}", ds) if "{{ ds }}" in file else file for file in file_list]

        kwargs['ti'].xcom_push(key='file_list', value=file_list)

    def file_sensor(self, table_config, taskgroup):
        """
        Returns an Airflow sensor that waits for the list of files specified by the metadata file provided.
        """
        # If we have a flag file for ingestion, there is no need to check the files specifically
        if "flag_file_path" in table_config.extra_options.get("file_table_config"):
            return None

        # bucket = self.config.source.extra_options["gcs_bucket"]
        table_name = table_config.table_name
        files_to_wait_for = "{{" + f"ti.xcom_pull(key='file_list', task_ids='{table_name}.ftp_taskgroup.get_file_list')" + "}}"
        timeout = self.config.source.extra_options["file_source_config"]["sensor_timeout"]
        sftp_conn = self.config.source.extra_options["sftp_source_config"]["sftp_connection_name"]
        file_prefix_filtering = timeout = self.config.source.extra_options["file_source_config"]["file_prefix_filtering"]
        dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]

        if file_prefix_filtering:
            if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
                key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
                return SFTPFilesExistencePrefixSensor(
                    task_id="wait_for_files_to_ingest",
                    path=dir_prefix,
                    prefixes=files_to_wait_for,
                    key_auth=True,
                    key_name=key_name,
                    task_group=taskgroup,
                    sftp_conn_id=sftp_conn,
                    timeout=timeout
                )
            else:
                return SFTPFilesExistencePrefixSensor(
                    task_id="wait_for_files_to_ingest",
                    path=dir_prefix,
                    prefixes=files_to_wait_for,
                    task_group=taskgroup,
                    sftp_conn_id=sftp_conn,
                    timeout=timeout
                )
        else:
            if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
                key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
                return SFTPFilesExistenceSensor(
                    task_id="wait_for_files_to_ingest",
                    dir_prefix=dir_prefix,
                    paths=files_to_wait_for,
                    key_auth=True,
                    key_name=key_name,
                    task_group=taskgroup,
                    sftp_conn_id=sftp_conn,
                    timeout=timeout
                )
            else:
                return SFTPFilesExistenceSensor(
                    task_id="wait_for_files_to_ingest",
                    dir_prefix=dir_prefix,
                    paths=files_to_wait_for,
                    task_group=taskgroup,
                    sftp_conn_id=sftp_conn,
                    timeout=timeout
                )

    def get_sftp_sensor(self, table_config, taskgroup, dir_prefix, flag_file_path):
        return PythonSensor(
            task_id="sftp_sensor",
            op_kwargs={"table_config": table_config,
                       "dir_prefix": dir_prefix,
                       "flag_file_path": flag_file_path},
            python_callable=self.sftp_flag_sensor,
            task_group=taskgroup
        )

    def sftp_flag_sensor(self, table_config, dir_prefix, flag_file_path, **kwargs):
        # Check if the flag file is dated with the execution date or later - if yes, then the files are ready for ingestion.
        # allows backfilling
        sftp_conn = self.config.source.extra_options["sftp_source_config"]["sftp_connection_name"]

        ds = kwargs["ds"]
        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            private_key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
            save_id_to_file(private_key_name)
        sftp_hook = SFTPHook(ssh_conn_id=sftp_conn)
        dir_prefix = dir_prefix.replace("{{ ds }}", ds)
        mod_time = sftp_hook.get_mod_time(flag_file_path)
        mod_time = datetime.strptime(mod_time, '%Y%m%d%H%M%S').strftime('%Y-%m-%d')
        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            if os.path.isfile("id_rsa"):
                os.remove("id_rsa")

        return ds <= mod_time

    def get_sftp_ingestion_operator(self, table_config, taskgroup, dir_prefix, gcs_bucket_prefix, flag_file_path, bucket):
        if not flag_file_path == "":
            return PythonOperator(
                task_id="load_sftp_to_gcs",
                op_kwargs={
                    "table_config": table_config,
                    "dir_prefix": dir_prefix,
                    "gcs_bucket_prefix": gcs_bucket_prefix,
                    "bucket": bucket,
                    "flag_file_path": flag_file_path},
                python_callable=self.load_sftp_to_gcs_flag,
                task_group=taskgroup
            )
        else:
            return PythonOperator(
                task_id="load_sftp_to_gcs",
                op_kwargs={
                    "table_config": table_config,
                    "dir_prefix": dir_prefix,
                    "gcs_bucket_prefix": gcs_bucket_prefix,
                    "bucket": bucket},
                python_callable=self.load_sftp_to_gcs,
                task_group=taskgroup
            )

    # flake8: noqa: C901
    def load_sftp_to_gcs(self, table_config, dir_prefix, gcs_bucket_prefix, bucket, **kwargs):
        ds = kwargs["ds"]
        ti = kwargs["ti"]
        sftp_conn = self.config.source.extra_options["sftp_source_config"]["sftp_connection_name"]

        if not os.path.isdir("sftp_data"):
            os.mkdir("sftp_data")
        my_path = "sftp_data"

        if gcs_bucket_prefix is None:
            gcs_bucket_prefix = ""
        if not gcs_bucket_prefix == "":
            gcs_bucket_prefix += "/"

        table_name = table_config.table_name

        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            private_key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
            save_id_to_file(private_key_name)
        sftp_hook = SFTPHook(sftp_conn)

        files_to_load = ti.xcom_pull(key='file_list', task_ids=f'{table_name}.ftp_taskgroup.get_file_list')

        if self.config.source.extra_options["file_source_config"]["file_prefix_filtering"]:
            file_name_prefixes = [f.split("/")[-1] for f in files_to_load]
            full_file_names = []
            dir = sftp_hook.list_directory(dir_prefix)
            for prefix in file_name_prefixes:
                for f in dir:
                    if f.startswith(prefix):
                        full_file_names.append(f)
                    break
            files_to_load = full_file_names

        for file in files_to_load:
            sftp_hook.retrieve_file(dir_prefix + "/" + file, my_path + "/" + file)

        destination_path_prefix = gcs_bucket_prefix + table_name + "/" + ds
        if "gcs_bucket_path_format_mode" in self.config.source.extra_options["file_source_config"]:
            date = datetime.strptime(ds, '%Y-%m-%d').strftime('%Y/%m/%d')
            destination_path_prefix = gcs_bucket_prefix + table_name + "/" + date

        gcs_hook = GCSHook()
        for file in files_to_load:
            obj = destination_path_prefix + "/" + file
            gcs_hook.upload(
                bucket_name=bucket,
                object_name=obj,
                filename=my_path + "/" + file)
        # delete data
        shutil.rmtree(my_path)

        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            if os.path.isfile("id_rsa"):
                os.remove("id_rsa")

    def load_sftp_to_gcs_flag(self, table_config, dir_prefix, gcs_bucket_prefix, bucket, flag_file_path, **kwargs):
        ds = kwargs["ds"]
        # ti = kwargs["ti"]

        airflow_date_template = self.config.source.extra_options["file_source_config"]["airflow_date_template"]

        if gcs_bucket_prefix is None:
            gcs_bucket_prefix = ""
        if not gcs_bucket_prefix == "":
            gcs_bucket_prefix += "/"

        # flag to check whether we are loading dates named with {{ prev_ds }} or {{ ds }}
        # landing zone table creation will still use {{ ds }}
        source_file_date = ds
        if airflow_date_template == "prev_ds":
            source_file_date = kwargs["prev_ds"]

        date_column = table_config.extra_options.get("sftp_table_config")["date_column"]
        full_dir_download = table_config.extra_options.get("sftp_table_config")["full_dir_download"]
        table_name = table_config.table_name
        destination_path_prefix = gcs_bucket_prefix + table_name + "/" + ds

        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            private_key_name = self.config.source.extra_options["sftp_source_config"]["sftp_private_key_secret_name"]
            save_id_to_file(private_key_name)
        sftp_hook = SFTPHook()

        # Create directory and download data
        if not os.path.isdir("sftp_data"):
            os.mkdir("sftp_data")
        my_path = "sftp_data"
        os.chmod(my_path, 0o777)

        # get recursive sftp structure w/ full paths
        tree_map_prefix = dir_prefix
        upload_option = table_config.extra_options.get("file_table_config")["parquet_upload_option"]
        if not full_dir_download and upload_option == "BASH":
            tree_map_prefix += f"/{date_column}={source_file_date}"
        logging.info(tree_map_prefix)
        map = sftp_hook.get_tree_map(tree_map_prefix)
        if isinstance(map[0], list):
            map = map[0]

        # get list of files, create local directory structure, download
        files_to_upload = [x for x in map]
        # delete prefix up until the external partitions
        local_file_prefix_to_drop = ""
        if not files_to_upload == []:
            local_file_prefix_to_drop = my_path + "/" + dir_prefix + "/"

        gcs_hook = GCSHook()
        logging.info(files_to_upload)

        local_files_list = [my_path + "/" + file for file in files_to_upload]

        for remote_file in files_to_upload:
            file = my_path + "/" + remote_file
            path_to_create = file.rsplit('/', 1)[0]
            if not os.path.exists(path_to_create):
                folders = path_to_create.split("/")
                for i in range(len(folders)):
                    folder_to_create = "/".join(folders[0:i + 1])
                    if not os.path.exists(folder_to_create):
                        os.mkdir(folder_to_create)
                        os.chmod(folder_to_create, 0o777)

            if not remote_file == flag_file_path:
                sftp_hook.retrieve_file(remote_file, file)
                # Get correct GCS file names to upload
                obj = destination_path_prefix + "/" + file.replace(local_file_prefix_to_drop, "")
                logging.info(obj)
                gcs_hook.upload(
                    bucket_name=bucket,
                    object_name=obj,
                    filename=file)
                os.remove(file)

        if "sftp_private_key_secret_name" in self.config.source.extra_options["sftp_source_config"]:
            if os.path.isfile("id_rsa"):
                os.remove("id_rsa")

        # xcom push the prefix of external partitions, e.g. "par1=val1/par2=val/par3=val3"
        # hacky way to do it - find a cleaner way to fix it later

        if "partition_prefix" not in self.config.source.extra_options["sftp_source_config"]:
            partition_prefix = local_files_list[0].rsplit('/', 1)[0].replace(my_path + "/" + dir_prefix, "").lstrip("/")
            if "/" in partition_prefix:
                partition_prefix = partition_prefix.rsplit('/', 1)[0]
        else:
            # TEMPORARY hack to fix ingestion. # TO REMOVE
            if table_config.extra_options.get("file_table_config")["parquet_upload_option"] == "BASH":
                partition_prefix = self.config.source.extra_options["sftp_source_config"]["partition_prefix"]
                partition_prefix.replace("date", table_config.extra_options.get("sftp_table_config")["date_column"])
                partition_prefix.replace("ds", ds)

        logging.info(partition_prefix)
        kwargs['ti'].xcom_push(key='partition_prefix', value=partition_prefix)

    def delete_gcs_files(self, table_config, taskgroup):
        pass

    def validate_extra_options(self):
        super().validate_extra_options()
        tables = self.config.tables
        for table_config in tables:
            # try and parse as SFTPTableConfig
            # sftp_table_config = from_dict(data_class=SFTPTableConfig, data=table_config.extra_options.get("sftp_table_config"))
            pass



def parse_ods_schema(ods_metadata, schema_fields) -> list:

    ODS_EXTRA_FIELDS = [
        {"name": ods_metadata.ingestion_time_column_name, "type": "TIMESTAMP"},
        {"name": ods_metadata.update_time_column_name, "type": "TIMESTAMP"},
        {"name": ods_metadata.primary_key_hash_column_name, "type": "STRING"},
        {"name": ods_metadata.hash_column_name, "type": "STRING"},
    ]

    ods_schema_fields = []

    ods_schema_fields.extend(schema_fields)
    ods_schema_fields.extend(ODS_EXTRA_FIELDS)

    return ods_schema_fields



import logging
from abc import abstractmethod
from datetime import datetime
import json
from dacite import from_dict
from os.path import join

from airflow.models.dag import DAG
from airflow.operators.python import PythonOperator
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import GCSToBigQueryOperator
from airflow.utils.task_group import TaskGroup
from airflow.operators.bash import BashOperator
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.sensors.gcs import GCSObjectExistenceSensor
from airflow.operators.dummy import DummyOperator
from gcp_airflow_foundations.base_class import file_source_config

from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.base_class.file_source_config import FileSourceConfig
from gcp_airflow_foundations.base_class.file_table_config import FileTableConfig


class GenericFileIngestionDagBuilder(DagBuilder):
    """
    Builds DAGs to load files from a generic file system to BigQuery.
    """
    source_type = "FTP"

    def set_schema_method_type(self):
        self.schema_source_type = self.config.source.schema_options.schema_source_type

    def get_bq_ingestion_task(self, dag, table_config):
        taskgroup = TaskGroup(group_id="ftp_taskgroup")
        file_source_config = from_dict(data_class=FileSourceConfig, data=self.config.source.extra_options["file_source_config"])

        tasks = []

        skip_gcs_upload = False
        if "skip_gcs_upload" in self.config.source.extra_options["file_source_config"]:
            skip_gcs_upload = True

        if not skip_gcs_upload:
            tasks.append(self.metadata_file_sensor(table_config, taskgroup))
            tasks.append(self.flag_file_sensor(table_config, taskgroup))
            tasks.append(self.schema_file_sensor(table_config, taskgroup))
            tasks.append(self.get_file_list_task(table_config, taskgroup))
            tasks.append(self.file_sensor(table_config, taskgroup))
            tasks.append(self.file_ingestion_task(table_config, taskgroup))

        tasks.append(self.load_to_landing_task(table_config, taskgroup))

        if file_source_config.delete_gcs_files:
            tasks.append(self.delete_gcs_files(table_config, taskgroup))

        for task in tasks:
            if task is None:
                tasks.remove(task)

        not_none_tasks = list(filter(None.__ne__, tasks))

        for i in range(len(not_none_tasks) - 1):
            not_none_tasks[i] >> not_none_tasks[i + 1]

        return taskgroup

    def metadata_file_sensor(self, table_config, taskgroup):
        """
        Implements a sensor for either the metadata file specified in the table config, which specifies
        the parameterized file names to ingest.
        """
        file_source_config = from_dict(data_class=FileSourceConfig, data=self.config.source.extra_options["file_source_config"])
        if "metadata_file" in table_config.extra_options.get("file_table_config"):
            metadata_file_name = table_config.extra_options.get("file_table_config")["metadata_file"]
            bucket = self.config.source.extra_options["gcs_bucket"]
            timeout = file_source_config.sensor_timeout

            return GCSObjectExistenceSensor(
                task_id="wait_for_metadata_file",
                bucket=bucket,
                object=metadata_file_name,
                task_group=taskgroup,
                timeout=timeout
            )
        else:
            return None

    @abstractmethod
    def flag_file_sensor(self, table_config):
        """
        Implements an Airflow sensor to wait for optional flag files for ingestion.
        e.g. for .PARQUET file ingestion, waiting for a _SUCCESS file is part of a common flow.
        """
        pass

    def schema_file_sensor(self, table_config, taskgroup):
        """
        Implements an Airflow sensor to wait for an (optional) schema file in GCS
        """
        file_source_config = from_dict(data_class=FileSourceConfig, data=self.config.source.extra_options["file_source_config"])
        bucket = self.config.source.extra_options["gcs_bucket"]
        schema_file_name = None
        timeout = file_source_config.sensor_timeout
        if "schema_file" in table_config.extra_options.get("file_table_config"):
            schema_file_name = table_config.extra_options.get("file_table_config")["schema_file"]
            return GCSObjectExistenceSensor(
                task_id="wait_for_schema_file",
                bucket=bucket,
                object=schema_file_name,
                task_group=taskgroup,
                timeout=timeout
            )
        else:
            return None

    @abstractmethod
    def file_ingestion_task(self, table_config):
        """
        Implements an Airflow task to ingest the files from the FTP source into GCS (e.g. from an SFTP server or an AWS bucket)
        """
        pass

    @abstractmethod
    def file_sensor(self, table_config):
        """
        Returns an Airflow sensor that waits for the list of files specified the metadata file provided
        Should be Xcom pulled from get_file_list_task()
        """
        pass

    @abstractmethod
    def delete_gcs_files(table_config, taskgroup):
        pass

    def get_file_list_task(self, table_config, taskgroup):
        return PythonOperator(
            task_id="get_file_list",
            op_kwargs={"table_config": table_config},
            python_callable=self.get_list_of_files,
            task_group=taskgroup
        )

    def get_list_of_files(self, table_config, **kwargs):
        # gcs_hook = GCSHook()
        file_source_config = from_dict(data_class=FileSourceConfig, data=self.config.source.extra_options["file_source_config"])
        airflow_date_template = file_source_config.airflow_date_template
        if airflow_date_template == "ds":
            ds = kwargs["ds"]
        else:
            ds = kwargs["prev_ds"]
        ds = datetime.strptime(ds, "%Y-%m-%d").strftime(file_source_config.date_format)
        logging.info(ds)
        # XCom push the list of files
        # overwrite if in table_config
        dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]
        dir_prefix = dir_prefix.replace("{{ ds }}", ds)

        gcs_bucket_prefix = file_source_config.gcs_bucket_prefix

        if file_source_config.source_format == "PARQUET":
            file_list = [dir_prefix]
            kwargs['ti'].xcom_push(key='file_list', value=file_list)
            return
        else:
            # bucket = self.config.source.extra_options["gcs_bucket"]
            if "metadata_file" in table_config.extra_options.get("file_table_config"):
                # metadata_file_name = table_config.extra_options.get("file_table_config")["metadata_file"]
                # metadata_file = gcs_hook.download(bucket_name=bucket, object_name=metadata_file_name, filename="metadata.csv")
                file_list = []
                with open('metadata.csv', newline='') as f:
                    for line in f:
                        file_list.append(line.strip())
            else:
                templated_file_name = file_source_config.file_name_template
                templated_file_name = templated_file_name.replace("{{ TABLE_NAME }}", table_config.table_name)
                file_list = [templated_file_name]

            # support replacing files with current dates
            file_list[:] = [file.replace("{{ ds }}", ds) if "{{ ds }}" in file else file for file in file_list]
            # add dir prefix to files
            file_list[:] = [join(gcs_bucket_prefix, file) for file in file_list]
            logging.info(file_list)

        kwargs['ti'].xcom_push(key='file_list', value=file_list)

    def load_to_landing_task(self, table_config, taskgroup):
        return PythonOperator(
            task_id="load_gcs_to_landing_zone",
            op_kwargs={"table_config": table_config},
            python_callable=self.load_to_landing,
            task_group=taskgroup
        )

    # flake8: noqa: C901
    def load_to_landing(self, table_config, **kwargs):
        gcs_hook = GCSHook()
        file_source_config = from_dict(data_class=FileSourceConfig, data=self.config.source.extra_options["file_source_config"])

        # Parameters
        ds = kwargs['ds']
        ti = kwargs['ti']

        data_source = self.config.source
        bucket = data_source.extra_options["gcs_bucket"]
        source_format = file_source_config.source_format
        field_delimeter = file_source_config.delimeter
        gcp_project = data_source.gcp_project
        landing_dataset = data_source.landing_zone_options.landing_zone_dataset
        landing_table_name = table_config.landing_zone_table_name_override
        table_name = table_config.table_name
        destination_table = f"{gcp_project}:{landing_dataset}.{table_config.landing_zone_table_name_override}" + f"_{ds}"

        if "skip_gcs_upload" not in data_source.extra_options["file_source_config"]:
            files_to_load = ti.xcom_pull(key='file_list', task_ids=f'{table_name}.ftp_taskgroup.get_file_list')
        else:
            dir_prefix = table_config.extra_options.get("file_table_config")["directory_prefix"]
            dir_prefix = dir_prefix.replace("{{ ds }}", ds)
            files_to_load = [dir_prefix]

        gcs_bucket_prefix = file_source_config.gcs_bucket_prefix
        if gcs_bucket_prefix is None:
            gcs_bucket_prefix = ""
        if not gcs_bucket_prefix == "":
            gcs_bucket_prefix += "/"

        destination_path_prefix = gcs_bucket_prefix + table_name + "/" + ds
        if "gcs_bucket_path_format_mode" in self.config.source.extra_options["file_source_config"]:
            date = datetime.strptime(ds, '%Y-%m-%d').strftime('%Y/%m/%d')
            destination_path_prefix = gcs_bucket_prefix + table_name + "/" + date
            logging.info(destination_path_prefix)

            files_to_load = [destination_path_prefix + "/" + f for f in files_to_load]
            logging.info(files_to_load)

        if "parquet_upload_option" in table_config.extra_options.get("file_table_config"):
            parquet_upload_option = table_config.extra_options.get("file_table_config")["parquet_upload_option"]
        else:
            parquet_upload_option = "BASH"

        source_format = file_source_config.source_format
        if source_format == "PARQUET" and parquet_upload_option == "BASH":
            date_column = table_config.extra_options.get("sftp_table_config")["date_column"]
            gcs_bucket_prefix = file_source_config.gcs_bucket_prefix
            # bq load command if parquet
            partition_prefix = ti.xcom_pull(key='partition_prefix', task_ids=f'{table_name}.ftp_taskgroup.load_sftp_to_gcs')
            if not partition_prefix:
                partition_prefix = self.config.source.extra_options["sftp_source_config"]["partition_prefix"]
                partition_prefix = partition_prefix.replace("date", table_config.extra_options.get("sftp_table_config")["date_column"])
                partition_prefix = partition_prefix.replace("ds", kwargs['prev_ds'])
            if "prefix" in table_config.extra_options.get("file_table_config"):
                partition_prefix = partition_prefix + "/" + table_config.extra_options.get("file_table_config")["prefix"]
            command = self.get_load_script(gcp_project, landing_dataset, landing_table_name + f"_{ds}", bucket, gcs_bucket_prefix, partition_prefix, table_name, date_column, ds)
            logging.info(command)
            try:
                bash = BashOperator(
                    task_id="import_files_to_bq_landing",
                    bash_command=command
                )
                bash.execute(context=kwargs)
            except Exception:
                logging.info("Load into BQ landing zone failed.")
        else:
            # gcs->bq operator else
            if file_source_config.file_prefix_filtering:
                for i in range(len(files_to_load)):
                    matching_gcs_files = gcs_hook.list(bucket_name=bucket, prefix=files_to_load[i])
                    logging.info(matching_gcs_files)
                    if len(matching_gcs_files) > 1:
                        raise AirflowException(f"There is more than one matching file with the prefix {files_to_load[i]} in the bucket {bucket}")
                    files_to_load[i] = matching_gcs_files[0]
            else:
                files_to_load = [f"{gcs_bucket_prefix}{table_name}/{ds}/" + f for f in files_to_load]

            schema_file_name = None
            if "schema_file" in table_config.extra_options.get("file_table_config"):
                schema_file_name = table_config.extra_options.get("file_table_config")["schema_file"]

            allow_quoted_newlines = False
            if "allow_quoted_newlines" in table_config.extra_options.get("file_table_config"):
                allow_quoted_newlines = table_config.extra_options.get("file_table_config")["allow_quoted_newlines"]

            if parquet_upload_option == "GCS" and source_format == "PARQUET":
                prefix = ""
                if "prefix" in table_config.extra_options.get("file_table_config"):
                    prefix = table_config.extra_options.get("file_table_config")["prefix"]
                prefix = destination_path_prefix + "/" + prefix
                # logging.info(destination_path_prefix + "/" + partition_prefix)
                files_to_load = gcs_hook.list(bucket_name=bucket, prefix=prefix)

            # Get files to load from metadata file
            if schema_file_name:
                schema_file = gcs_hook.download(bucket_name=bucket, object_name=schema_file_name)
                # Only supports json schema file format - add additional support if required
                schema_fields = json.loads(schema_file)
                gcs_to_bq = GCSToBigQueryOperator(
                    task_id='import_files_to_bq_landing',
                    bucket=bucket,
                    source_objects=files_to_load,
                    source_format=source_format,
                    schema_fields=schema_fields,
                    field_delimiter=field_delimeter,
                    destination_project_dataset_table=destination_table,
                    allow_quoted_newlines=allow_quoted_newlines,
                    write_disposition='WRITE_TRUNCATE',
                    create_disposition='CREATE_IF_NEEDED',
                    skip_leading_rows=1,
                )
            else:
                gcs_to_bq = GCSToBigQueryOperator(
                    task_id='import_files_to_bq_landing',
                    bucket=bucket,
                    source_objects=files_to_load,
                    source_format=source_format,
                    field_delimiter=field_delimeter,
                    destination_project_dataset_table=destination_table,
                    allow_quoted_newlines=allow_quoted_newlines,
                    write_disposition='WRITE_TRUNCATE',
                    create_disposition='CREATE_IF_NEEDED',
                    skip_leading_rows=1,
                )
            gcs_to_bq.execute(context=kwargs)

            kwargs['ti'].xcom_push(key='loaded_files', value=files_to_load)

    def get_load_script(self, gcp_project, landing_dataset, landing_table_name, bucket, gcs_bucket_prefix, partition_prefix, table_name, date_column, ds):
        if not partition_prefix == "":
            partition_prefix += "/"
        full_table_name = f"{landing_dataset}.{landing_table_name}"
        source_uri_prefix = f"gs://{bucket}/{gcs_bucket_prefix}{table_name}/{ds}"
        uri_wildcards = f"gs://{bucket}/{gcs_bucket_prefix}{table_name}/{ds}/{partition_prefix}*"
        command = f"bq load --source_format=PARQUET --autodetect --hive_partitioning_mode=STRINGS --hive_partitioning_source_uri_prefix={source_uri_prefix} {full_table_name} {uri_wildcards}"
        logging.info(command)
        return command

    def validate_extra_options(self):
        # try and parse as FTPSourceConfig
        # file_source_config = from_dict(data_class=FileSourceConfig, data=self.config.source.extra_options["file_source_config"])

        tables = self.config.tables
        for table_config in tables:
            # try and parse as FTPTableConfig
            # file_table_config = from_dict(data_class=FileTableConfig, data=table_config.extra_options.get("file_table_config"))
            pass



from typing import Optional

from airflow.providers.google.cloud.operators.bigquery import (
    BigQueryDeleteTableOperator,
)

# from airflow.utils.decorators import apply_defaults

from airflow.exceptions import AirflowException

import logging


class BigQueryDeleteStagingTableOperator(BigQueryDeleteTableOperator):
    # @apply_defaults
    def __init__(
        self,
        *,
        project_id: str,
        dataset_id: str,
        table_id: str,
        gcp_conn_id: str = "google_cloud_default",
        **kwargs,
    ) -> None:

        deletion_dataset_table = f"{project_id}.{dataset_id}.{table_id}"

        super(BigQueryDeleteStagingTableOperator, self).__init__(
            deletion_dataset_table=deletion_dataset_table,
            ignore_if_missing=True,
            **kwargs,
        )

    def pre_execute(self, context) -> None:
        ds = context["ds"]
        self.deletion_dataset_table = f"{self.deletion_dataset_table}_{ds}"



from pydantic import validator
from pydantic.dataclasses import dataclass

from typing import List, Optional


@dataclass
class HdsTableMetadataConfig:
    """
    Attributes:
        eff_start_time_column_name : Insertion time column name
        eff_end_time_column_name: Expiry time column name
        partition_time_column_name: Partition time column name
        hash_column_name : Hash column name
    """

    eff_start_time_column_name: str = "af_metadata_created_at"
    eff_end_time_column_name: str = "af_metadata_expired_at"
    partition_time_column_name: str = "partition_time"
    hash_column_name: str = "af_metadata_row_hash"



import datetime

from airflow.models.dag import DAG
from airflow.models import Variable
from airflow.exceptions import AirflowException
from gcp_airflow_foundations.base_class.utils import load_tables_config_from_dir
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.source_class import get_dag_builder
from gcp_airflow_foundations.source_class import (
    generic_file_source,
    gcs_source,
    sftp_source,
    jdbc_dataflow_source,
    oracle_dataflow_source,
    salesforce_source,
    twilio_source,
    mysql_dataflow_source,
)

import logging


class DagParser:
    """ Parsed Airflow DAGs from the user-provided YAML configuration files """

    def __init__(self):
        logging.info("Initialize DagParser")
        self.conf_location = Variable.get("CONFIG_FILE_LOCATION", "dags/config")

    def parse_dags(self):
        logging.info(f"Loading config from dir: {self.conf_location}")

        configs = load_tables_config_from_dir(self.conf_location)

        parsed_dags = {}
        for config in configs:
            logging.info(
                f"StartDate for {config.source.name}: {config.source_start_date()}"
            )

            default_task_args = {
                "owner": config.source.owner,
                "retries": config.source.num_retries,
                "retry_exponential_backoff": True,
                "retry_delay": datetime.timedelta(seconds=300),
                "project_id": config.source.gcp_project,
                "email": config.source.notification_emails,
                "depends_on_past": False,
                "email_on_retry": config.source.email_on_retry,
                "email_on_failure": config.source.email_on_failure
            }

            """
            # Old way of adding DAGs
            builder = get_dag_builder(config.source.source_type, default_task_args, config)
            dags = builder.build_dags()

            for dag in dags:
                parsed_dags[f"dags:source:{config.source.name}.{dag.dag_id}"] = dag
            """

            # new way
            builder = None
            dags = []
            for dag_builder in DagBuilder.sources:
                # if matching subclass of DagBuilder exists, then use it
                if config.source.source_type == dag_builder.source_type:
                    # source_name = dag_builder.__class__.__name__
                    # pick out the right source
                    builder = dag_builder(default_task_args, config)
                    dags = builder.build_dags()

            if builder is None:
                raise AirflowException(
                    f'Source "{config.source.source_type}" is not found in DagBuilder Class'
                )

            for dag in dags:
                parsed_dags[f"dags:source:{config.source.name}.{dag.dag_id}"] = dag

        return parsed_dags



from typing import Optional
from datetime import datetime

from airflow.models import BaseOperator, BaseOperatorLink
from airflow.contrib.operators.bigquery_operator import (
    BigQueryOperator,
    BigQueryCreateEmptyTableOperator,
)

from airflow.utils.decorators import apply_defaults
from airflow.contrib.hooks.bigquery_hook import BigQueryHook

from airflow.exceptions import AirflowException

import logging

from gcp_airflow_foundations.operators.gcp.ods.ods_sql_upsert_helpers import (
    SqlHelperODS,
)
from gcp_airflow_foundations.base_class.ods_table_config import OdsTableConfig
from gcp_airflow_foundations.enums.ingestion_type import IngestionType


class MergeBigQueryODS(BigQueryOperator):
    """
    Merges data into a BigQuery ODS table.

    :param project_id: GCP project ID
    :type project_id: str
    :param stg_table_name: Source table name
    :type stg_table_name: str
    :param data_table_name: Target table name
    :type data_table_name: str
    :param stg_dataset_name: Source dataset name
    :type stg_dataset_name: str
    :param data_dataset_name: Target dataset name
    :type data_dataset_name: str
    :param surrogate_keys: Surrogate keys of the target table
    :type surrogate_keys: list
    :param delegate_to: The account to impersonate using domain-wide delegation of authority, if any
    :type delegate_to: str
    :param gcp_conn_id: Airflow GCP connection ID
    :type gcp_conn_id: str
    :param column_mapping: Column mapping
    :type column_mapping: dict
    :param new_column_udfs: New column UDFs
    :type new_column_udfs: dict
    :param ingestion_type: Source table ingestion time (Full or Incremental)
    :type ingestion_type: IngestionType
    :param ods_table_config: User-provided ODS configuration options
    :type ods_table_config: OdsTableConfig
    :param location: The geographic location of the job.
    :type location: str
    """

    template_fields = ("stg_table_name", "data_table_name", "stg_dataset_name")

    @apply_defaults
    def __init__(
        self,
        *,
        project_id: str,
        stg_table_name: str,
        data_table_name: str,
        dag_table_id: str,
        stg_dataset_name: str,
        data_dataset_name: str,
        surrogate_keys: [str],
        columns: Optional[list] = None,
        delegate_to: Optional[str] = None,
        gcp_conn_id: str = "google_cloud_default",
        column_mapping: dict,
        ingestion_type: IngestionType,
        ods_table_config: OdsTableConfig,
        column_casting: dict = {},
        new_column_udfs: dict = {},
        location: Optional[str] = None,
        **kwargs,
    ) -> None:
        super(MergeBigQueryODS, self).__init__(
            delegate_to=delegate_to,
            gcp_conn_id=gcp_conn_id,
            use_legacy_sql=False,
            write_disposition="WRITE_APPEND",
            create_disposition="CREATE_NEVER",
            location=location,
            sql="",
            **kwargs,
        )
        self.project_id = project_id
        self.stg_table_name = stg_table_name
        self.data_table_name = data_table_name
        self.dag_table_id = dag_table_id
        self.stg_dataset_name = stg_dataset_name
        self.data_dataset_name = data_dataset_name
        self.surrogate_keys = surrogate_keys
        self.columns = columns
        self.gcp_conn_id = gcp_conn_id
        self.delegate_to = delegate_to
        self.column_mapping = column_mapping
        self.column_casting = column_casting
        self.new_column_udfs = new_column_udfs
        self.ingestion_type = ingestion_type
        self.ods_table_config = ods_table_config

    def pre_execute(self, context) -> None:
        ds = context["ds"]

        if not self.columns:
            staging_columns = self.xcom_pull(
                context=context, task_ids=f"{self.dag_table_id}.schema_parsing"
            )["source_table_columns"]
            logging.info(staging_columns)
        else:
            staging_columns = self.columns

        if self.column_mapping:
            for i in staging_columns:
                if i not in self.column_mapping:
                    self.column_mapping[i] = i

        else:
            self.column_mapping = {i: i for i in staging_columns}

        source_columns = staging_columns

        self.log.info(
            f"Execute BigQueryMergeTableOperator {self.stg_table_name}, {self.data_table_name}"
        )

        sql = ""

        sql_helper = SqlHelperODS(
            source_dataset=f"{self.project_id}.{self.stg_dataset_name}",
            target_dataset=f"{self.project_id}.{self.data_dataset_name}",
            source=f"{self.stg_table_name}_{ds}",
            target=self.data_table_name,
            columns=source_columns,
            surrogate_keys=self.surrogate_keys,
            column_mapping=self.column_mapping,
            column_casting=self.column_casting,
            new_column_udfs=self.new_column_udfs,
            ods_metadata=self.ods_table_config.ods_metadata,
        )

        if self.ods_table_config.ods_table_time_partitioning:
            partitioning_dimension = (
                self.ods_table_config.ods_table_time_partitioning.value
            )
            sql_helper.time_partitioning = partitioning_dimension
            sql_helper.partition_column_name = (
                self.ods_table_config.partition_column_name
            )

            ts = context["ts"]

            sql_helper.partition_timestamp = ts

        if self.ingestion_type == IngestionType.INCREMENTAL:
            if self.surrogate_keys:
                # Append staging table to ODS table
                sql = sql_helper.create_upsert_sql_with_hash()
            else:
                sql = sql_helper.create_full_sql()
                self.write_disposition = "WRITE_APPEND"
                self.destination_dataset_table = (
                    f"{self.project_id}.{self.data_dataset_name}.{self.data_table_name}"
                )

        elif self.ingestion_type == IngestionType.FULL:
            # Overwrite ODS table with the staging table data
            sql = sql_helper.create_full_sql()
            self.write_disposition = "WRITE_TRUNCATE"
            self.destination_dataset_table = (
                f"{self.project_id}.{self.data_dataset_name}.{self.data_table_name}"
            )

        else:
            raise AirflowException("Invalid ingestion type", self.ingestion_type)

        logging.info(
            f"Executing sql: {sql}. Write disposition: {self.write_disposition}"
        )

        self.sql = sql
        logging.info(sql)


import json
import logging

from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook


def read_schema_from_bq(
    project_id,
    dataset_id,
    table_id,
    google_cloud_storage_conn_id="google_cloud_default",
    gcp_conn_id="google_cloud_default",
    **kwargs
) -> list:
    """
        Helper method to load table schema from the staging table
    """

    bq_hook = BigQueryHook(gcp_conn_id=gcp_conn_id, delegate_to=None)

    schema = bq_hook.get_schema(dataset_id=dataset_id, table_id=table_id, project_id=project_id)

    return schema["fields"]



from airflow.exceptions import AirflowException

from dacite import Config
from dataclasses import dataclass, field

from datetime import datetime
from typing import List, Optional


@dataclass
class SalesforceIngestionConfig:
    """
    Salesforce configuration class.

    Attributes:
        ingest_all_columns: SELECT * the Salesforce object if true
        fields_to_omit: a list of object fields to omit from ingestion
        field_names: an explicit list of fields to ingest
    """

    api_table_name: str
    ingest_all_columns: bool
    fields_to_omit: Optional[List[str]]
    field_names: Optional[List[str]]



from dataclasses import fields
from urllib.parse import urlparse
import logging

from airflow.models.dag import DAG
from airflow.operators.python_operator import PythonOperator
from airflow.providers.google.cloud.transfers.gcs_to_bigquery import (
    GCSToBigQueryOperator,
)
from airflow.sensors.external_task import ExternalTaskSensor
from airflow.providers.google.cloud.hooks.kms import CloudKMSHook
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook
from airflow.models import Variable

from gcp_airflow_foundations.source_class.jdbc_dataflow_source import (
    JdbcToBQDataflowDagBuilder,
)
from gcp_airflow_foundations.base_class.data_source_table_config import (
    DataSourceTablesConfig,
)
from gcp_airflow_foundations.source_class.source import DagBuilder
from gcp_airflow_foundations.common.gcp.load_builder import load_builder
import gcp_airflow_foundations.common.dataflow.jdbc.oracle.oracle_query_helpers as oracle_helpers


class OracleToBQDataflowDagBuilder(JdbcToBQDataflowDagBuilder):
    """
    Builds DAGs to load a CSV file from GCS to a BigQuery Table.
    """

    source_type = "ORACLE"

    def create_job_params(
        self,
        config_params,
        destination_table,
        table_name,
        destination_schema_table,
        query_schema,
        owner,
        ingestion_type,
        table_params,
        **kwargs,
    ):
        #   1.  Generate SQL Query
        #    a. get source schema from BQ table
        bq_hook = BigQueryHook()
        if not query_schema:
            schema_query = oracle_helpers.get_table_schema_query(
                destination_schema_table, table_name
            )
            schema_df = bq_hook.get_pandas_df(schema_query, dialect="standard")

            column_names = schema_df.iloc[:, 0]
            data_types = schema_df.iloc[:, 1]

            #    b. apply castings if provided
            casted_columns = oracle_helpers.cast_columns(
                column_names, data_types, config_params["sql_casts"]
            )
            logging.info(casted_columns)

            #    c. get query
            if ingestion_type == "FULL":
                query = oracle_helpers.get_query_for_oracle_load_full(
                    table_name, casted_columns, owner
                )
            else:
                query = oracle_helpers.get_query_for_oracle_load_full(
                    table_name, casted_columns, owner
                )
                #query = oracle_helpers.get_query_for_oracle_load_incremental(
                #    table_name, casted_columns, table_params["date_column"], owner, kwargs["ds"], table_params["num_backtrack_days"]
                #)
            logging.info(query)

        else:
            query = oracle_helpers.get_schema_query(owner)

        # 2. KMS-encrypt credentials for Dataflow job
        logging.info("Encrypting DB credentials with Google Cloud KMS.")

        kms_hook = CloudKMSHook()
        kms_key = config_params["kms_key_path"]

        jdbc_url = bytes(config_params["jdbc_url"], "utf-8")
        jdbc_user = bytes(config_params["jdbc_user"], "utf-8")
        jdbc_pass = bytes(Variable.get(config_params["jdbc_pass_secret_name"]), "utf-8")

        enc_url = kms_hook.encrypt(key_name=kms_key, plaintext=jdbc_url)
        enc_user = kms_hook.encrypt(key_name=kms_key, plaintext=jdbc_user)
        enc_pass = kms_hook.encrypt(key_name=kms_key, plaintext=jdbc_pass)

        # Set up parameters to pass to Dataflow job, and XCom push them
        dataflow_default_options = {
            "project": config_params["project"],
            # "region": config_params["region"],
            "subnetwork": config_params["subnetwork"],
        }
        logging.info(dataflow_default_options)

        ds = kwargs["ds"]

        parameters = {
            "driverClassName": config_params["jdbc_driver_class"],
            "driverJars": config_params["jdbc_jar_path"],
            "KMSEncryptionKey": kms_key,
            "outputTable": destination_table + f"_{ds}",
            "bigQueryLoadingTemporaryDirectory": config_params[
                "bq_load_temp_directory"
            ],
            "connectionURL": enc_url,
            "username": enc_user,
            "password": enc_pass,
            "query": query,
        }
        logging.info(parameters)

        kwargs["ti"].xcom_push(
            key="dataflow_default_options", value=dataflow_default_options
        )
        kwargs["ti"].xcom_push(key="parameters", value=parameters)

    def get_landing_schema(self, schema_table, source_table):

        bq_hook = BigQueryHook()

        schema_query = oracle_helpers.get_table_schema_query(schema_table, source_table)
        schema_df = bq_hook.get_pandas_df(schema_query, dialect="standard")
        schema_df = schema_df.values.tolist()

        column_names = [x[0] for x in schema_df]
        dtypes = [x[1] for x in schema_df]

        dtypes = oracle_helpers.oracle_to_bq(dtypes)

        schema_fields = []
        for i in range(len(dtypes)):
            schema_fields.append(
                {"name": column_names[i], "type": dtypes[i], "mode": "NULLABLE"}
            )

        logging.info(schema_fields)

        return schema_fields



from airflow.exceptions import AirflowException

from dataclasses import dataclass
from dacite import from_dict

from pydantic import root_validator

from regex import E

from gcp_airflow_foundations.enums.hds_table_type import HdsTableType
from gcp_airflow_foundations.base_class.column_udf_config import ColumnUDFConfig


@dataclass
class SourceBaseConfig:
    """
    Base table/template configuration data class.
    """

    @root_validator(pre=True)
    def valid_hds_dataset(cls, values):
        if (values["cluster_fields"] is not None) and (
            values["column_mapping"] is not None
        ):
            values["cluster_fields"] = [
                values["column_mapping"].get(field, field)
                for field in values["cluster_fields"]
            ]
        return values

    @root_validator(pre=True)
    def valid_column_casting(cls, values):
        if values["column_casting"] is not None:
            assert all(
                [
                    key not in values["surrogate_keys"]
                    for key in values["column_casting"]
                ]
            ), "Column casting is available only for non-key columns."

            if values["hds_config"] is not None:
                assert (
                    values["hds_config"].hds_table_type != HdsTableType.SCD2
                ), "Column casting is not currently supported for HDS SCD2 tables."
        return values

    @root_validator(pre=True)
    def valid_new_column_udfs(cls, values):
        if values["new_column_udfs"] is not None:
            assert all(
                [from_dict(data_class=ColumnUDFConfig, data=values["new_column_udfs"][col]) for col in values["new_column_udfs"].keys()]
            ), "New column UDFs must only contain 'function' and 'output_type' keys with corresponding values."

            if values["hds_config"] is not None:
                assert (
                    values["hds_config"].hds_table_type != HdsTableType.SCD2
                ), "New column UDFs is not currently supported for HDS SCD2 tables."
        return values
